from django.conf import settings
from django.contrib.auth import get_user_model
from djangoldp_project.models import Member
from djangoldp.check_integrity import is_alive
from urllib.parse import urlparse

def add_arguments(parser):
  parser.add_argument(
    "--ignore-project-members",
    default=False,
    nargs="?",
    const=True,
    help="Ignore project members related check",
  )

def check_integrity(options):
  print('---')
  print("DjangoLDP Project")
  if(not options["ignore_project_members"]):

    ignored = set()
    if(options["ignore"]):
      for target in options["ignore"].split(","):
        ignored.add(urlparse(target).netloc)

    resources_404 = set()
    resources_servers_offline = set()
    resources_map = dict()

    for obj in Member.objects.all():
      resources_map[obj.urlid] = obj
      if hasattr(obj, "user"):
        if(obj.user.urlid):
          if(not obj.user.urlid.startswith(settings.BASE_URL)):
            url = urlparse(obj.user.urlid).netloc
            if(url not in ignored):
              try:
                if(is_alive(obj.user.urlid, 404)):
                  resources_404.add(obj.urlid)
              except:
                resources_servers_offline.add(obj.urlid)
        else:
          resources_404.add(obj.urlid)
      if hasattr(obj, "project"):
        if(obj.project.urlid):
          if(not obj.project.urlid.startswith(settings.BASE_URL)):
            url = urlparse(obj.project.urlid).netloc
            if(url not in ignored):
              try:
                if(is_alive(obj.project.urlid, 404)):
                  resources_404.add(obj.urlid)
              except:
                resources_servers_offline.add(obj.urlid)
        else:
          resources_404.add(obj.urlid)

    if(len(resources_404) > 0):
      print("Faulted project memberships, 404:")
      for resource in resources_404:
        print("- "+resource)
      if(options["fix_404_resources"]):
        for resource in resources_404:
          try:
            resources_map[resource].delete()
          except:
            pass
        print("Fixed 404 project memberships")
      else:
        print("Fix them with `./manage.py check_integrity --fix-404-resources`")

    if(len(resources_servers_offline) > 0):
      print("Faulted project memberships, servers offline:")
      for resource in resources_servers_offline:
        print("- "+resource)
      if(options["fix_offline_servers"]):
        for resource in resources_servers_offline:
          try:
            resources_map[resource].delete()
          except:
            pass
        print("Fixed project memberships on offline servers")
      else:
        print("Fix them with `./manage.py check_integrity --fix-offline-servers`")

    if(len(resources_servers_offline) == 0 and len(resources_404) == 0):
      print('Everything goes fine')

  else:
    print("Ignoring djangoldp-project checks")


# Generated by Django 2.2.18 on 2021-05-20 18:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp_project', '0019_auto_20210519_1349'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='customer',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='projects', to='djangoldp_project.Customer'),
        ),
    ]



# Generated by Django 2.2.23 on 2021-05-19 11:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp_project', '0018_auto_20210504_1748'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='city',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name='customer',
            name='country',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AddField(
            model_name='customer',
            name='postcode',
            field=models.CharField(blank=True, max_length=10, null=True),
        ),
        migrations.AlterField(
            model_name='customer',
            name='address',
            field=models.TextField(blank=True, null=True),
        ),
    ]



from django.conf import settings

from djangoldp.permissions import LDPPermissions
from djangoldp_project.xmpp import XMPP_SERVERS, get_client_ip
from djangoldp_project.filters import CustomerFilterBackend, ProjectFilterBackend, ProjectMemberFilterBackend


class CustomerPermissions(LDPPermissions):
    filter_backends=[CustomerFilterBackend]

    def get_object_permissions(self, request, view, obj):
        from djangoldp_project.models import Member

        perms = super().get_object_permissions(request, view, obj)

        user = request.user
        if not user.is_anonymous:
            # members of one of their projects can view the customer
            if Member.objects.filter(project__customer=obj, user=user).exists():
                perms.add('view')

        return perms


class ProjectPermissions(LDPPermissions):
    filter_backends=[ProjectFilterBackend]

    def get_object_permissions(self, request, view, obj):
        perms = super().get_object_permissions(request, view, obj)

        user = request.user
        if not user.is_anonymous:
            # permissions gained by being a member, and admin
            if obj.members.filter(user=user).exists():
                perms = perms.union({'view'})

                if obj.members.filter(user=user).get().is_admin:
                    perms = perms.union({'add', 'change', 'delete'})

            # permissions gained by the project being public
            if obj.status == 'Public':
                perms = perms.union({'view', 'add'})

        return perms

    def get_container_permissions(self, request, view, obj=None, bypass=False):
        if obj is None:
            perms = super().get_container_permissions(request, view, obj)

            if not request.user.is_anonymous and not bypass:
                default_perms = getattr(settings, 'USER_AUTHENTICATED_PROJECT_PERMISSIONS', ['view', 'add'])
                perms = perms.union(set(default_perms))
        else:
            return self.get_object_permissions(request, view, obj)

        return perms
    
    def has_permission(self, request, view):
        if get_client_ip(request) in XMPP_SERVERS:
            return True

        return super().has_permission(request, view)
    
    def has_object_permission(self, request, view, obj):
        if get_client_ip(request) in XMPP_SERVERS:
            return True

        return super().has_object_permission(request, view, obj)


class ProjectMemberPermissions(LDPPermissions):
    filter_backends = [ProjectMemberFilterBackend]

    def get_object_permissions(self, request, view, obj):
        perms = super().get_object_permissions(request, view, obj)

        user = request.user

        if not user.is_anonymous:
            # the operation is on myself
            if obj.user == user:
                perms.add('view')

                if not obj.is_admin or obj.project.members.filter(is_admin=True).count() > 1:
                    perms.add('delete')

                if obj.project.status == 'Public':
                    perms = perms.union({'add', 'delete'})

            # the operation is on another member
            else:
                # permissions gained in public projects
                if obj.project.status == 'Public':
                    perms = perms.union({'view', 'add'})

                # permissions gained for all members
                if obj.project.members.filter(user=user).exists():
                    perms = perms.union({'view', 'add'})

                    # permissions gained for admins (on other users)
                    if obj.project.members.filter(user=user).get().is_admin:
                        perms = perms.union({'view', 'add'})

                        if not obj.is_admin:
                            perms.add('delete')
                    else:
                        perms.add('view')

        return perms
    
    def has_permission(self, request, view):
        if get_client_ip(request) in XMPP_SERVERS:
            return True

        return super().has_permission(request, view)
    
    def has_object_permission(self, request, view, obj):
        if get_client_ip(request) in XMPP_SERVERS:
            return True

        return super().has_object_permission(request, view, obj)



import uuid
import json

from rest_framework.test import APITestCase, APIClient

from djangoldp_project.permissions import ProjectMemberPermissions
from djangoldp_project.models import Project, Member, Customer
from djangoldp_project.tests.models import User


class PermissionsTestCase(APITestCase):
    def setUp(self):
        self.client = APIClient()

    def setUpLoggedInUser(self, is_superuser=False):
        self.user = User(email='test@mactest.co.uk', first_name='Test', last_name='Mactest', username='test',
                         password='glass onion', is_superuser=is_superuser)
        self.user.save()
        self.client.force_authenticate(user=self.user)

    def _get_random_project(self, status_choice='Public', customer=None):
        return Project.objects.create(name='Test', status=status_choice, customer=customer)

    def _get_random_customer(self, owner=None):
        return Customer.objects.create(owner=owner)

    def setUpProject(self, status_choice='Public'):
        self.project = self._get_random_project(status_choice)

    def _get_request_json(self, **kwargs):
        res = {
            '@context': {
                '@vocab': "http://happy-dev.fr/owl/#",
                'rdf': "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                'rdfs': "http://www.w3.org/2000/01/rdf-schema#",
                'ldp': "http://www.w3.org/ns/ldp#",
                'foaf': "http://xmlns.com/foaf/0.1/",
                'name': "rdfs:label",
                'acl': "http://www.w3.org/ns/auth/acl#",
                'permissions': "acl:accessControl",
                'mode': "acl:mode",
                'inbox': "http://happy-dev.fr/owl/#inbox",
                'object': "http://happy-dev.fr/owl/#object",
                'author': "http://happy-dev.fr/owl/#author",
                'account': "http://happy-dev.fr/owl/#account",
                'jabberID': "foaf:jabberID",
                'picture': "foaf:depiction",
                'firstName': "http://happy-dev.fr/owl/#first_name",
                'lastName': "http://happy-dev.fr/owl/#last_name",
                'isAdmin': "http://happy-dev.fr/owl/#is_admin"
            }
        }

        for kwarg in kwargs:
            if isinstance(kwargs[kwarg], str):
                res.update({kwarg: {'@id': kwargs[kwarg]}})
            else:
                res.update({kwarg: kwargs[kwarg]})

        return res

    def _get_random_user(self):
        return User.objects.create(email='{}@test.co.uk'.format(str(uuid.uuid4())), first_name='Test', last_name='Test',
                                   username=str(uuid.uuid4()))

    # test project permissions
    def test_list_project_anonymous(self):
        self.setUpProject('Public')
        response = self.client.get('/projects/')
        self.assertEqual(response.status_code, 403)

    def test_list_project_authenticated(self):
        self.setUpLoggedInUser()
        # a public project, a private project I'm in and a private project I'm not in
        another_user = self._get_random_user()
        public_project = self._get_random_project('Public')
        my_project = self._get_random_project('Private')
        Member.objects.create(project=my_project, user=self.user)
        private_project = self._get_random_project('Private')

        response = self.client.get('/projects/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 2)

    # test customer permissions
    def test_list_customer_anonymous(self):
        self._get_random_customer()
        response = self.client.get('/customers/')
        self.assertEqual(response.status_code, 403)
        # self.assertEqual(len(response.data['ldp:contains']), 1)

    def test_list_customer_authenticated(self):
        self.setUpLoggedInUser()
        self._get_random_customer()

        response = self.client.get('/customers/')
        self.assertEqual(response.status_code, 200)

    def test_get_customer_anonymous(self):
        customer = self._get_random_customer()

        response = self.client.get('/customers/{}/'.format(customer.pk))
        self.assertEqual(response.status_code, 403)

    def test_get_customer_authenticated(self):
        self.setUpLoggedInUser()
        customer = self._get_random_customer()

        response = self.client.get('/customers/{}/'.format(customer.pk))
        self.assertEqual(response.status_code, 404)

    def test_get_customer_owner(self):
        self.setUpLoggedInUser()
        customer = self._get_random_customer(owner=self.user)

        response = self.client.get('/customers/{}/'.format(customer.pk))
        self.assertEqual(response.status_code, 200)

    # members of one of their projects can view the customer
    def test_get_customer_project_member(self):
        self.setUpLoggedInUser()
        customer = self._get_random_customer()

        project = self._get_random_project(status_choice='Private', customer=customer)
        Member.objects.create(project=project, user=self.user)

        response = self.client.get('/customers/{}/'.format(customer.pk))
        self.assertEqual(response.status_code, 200)

    def test_post_customer_anonymous(self):
        response = self.client.post('/customers/', data=json.dumps({}), content_type='application/ld+json')
        self.assertEqual(response.status_code, 403)

    def test_post_customer_authenticated(self):
        self.setUpLoggedInUser()
        response = self.client.post('/customers/', data=json.dumps({}), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

    # removing a Member - I am an admin
    def test_delete_project_member_admin(self):
        self.setUpLoggedInUser()
        self.setUpProject('Private')

        Member.objects.create(project=self.project, user=self.user, is_admin=True)
        another_user = self._get_random_user()
        other = Member.objects.create(project=self.project, user=another_user, is_admin=False)

        self.assertEqual(Member.objects.count(), 2)
        response = self.client.delete('/project-members/{}/'.format(other.pk))
        self.assertEqual(response.status_code, 204)
        self.assertEqual(Member.objects.count(), 1)

    # removing self as Member - I am a superuser and an admin, but the last admin (403)
    def test_delete_self_from_project_superuser_admin(self):
        self.setUpLoggedInUser(is_superuser=True)
        self.setUpProject('Private')

        ProjectMemberPermissions.superuser_perms = []

        me = Member.objects.create(project=self.project, user=self.user, is_admin=True)
        response = self.client.delete('/project-members/{}/'.format(me.pk))
        self.assertEqual(response.status_code, 403)
